/*
Marcin Owsiany 8:00
4 7.12.2000
14 18
*/

#include <sys/types.h>
#include <sys/shm.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/msg.h>
#include <stdlib.h>
#include <errno.h>
#include "myshm.h"

#include <stdio.h>

extern int errno;

/* semafor:
 *	0 - blokada kolejki :	1 - zablokowana
 *				0 - wolna do uzycia
 *	1 - komunikaty:		0 - brak
 *				n - n komunikatow
 */
 
static struct sembuf s_lock[] = {
{	0, 0, 0		},	/* czekaj az kolejka bedzie wolna do uzycia */
{	0, 1, 0		}	/* i zablokuj ja */
};

static struct sembuf s_unlock[] = {
{	0, -1, 0 }	/* zwolnij kolejke */
};

static struct sembuf s_increlease[] = {
{	1,  1, 0 },	/* dodaj komunikat */
{	0, -1, 0 }	/* i odblokuj kolejke */
};

static struct sembuf s_decrelease[] = {
{	1, -1, 0	},	/* odejmij jeden komunikat */
{	0, -1, 0	}	/* i zwolnij kolejke */
};

static struct sembuf s_wait4[] = {
{	1, -1, 0	},	/* czekaj az bedzie komunikat, usun jeden */
{	0, 0, 0		},	/* i kolejka bedzie wolna */
{	0, 1, 0		}	/* wtedy zablokuj kolejke */
};

int sem_block_queue(key_t qkey)
{
	int s;
	
	if ((s = semget(qkey, 2, IPC_CREAT | 0666)) == -1)
		return -1;
	return semop(s, s_lock, 2);
}

int sem_inc_msg_release_queue(key_t qkey)
{
	int s;
	
	if ((s = semget(qkey, 2, 0666)) == -1)
		return -1;
	return semop(s, s_increlease, 2);
}

int sem_dec_msg_release_queue(key_t qkey)
{
	int s;
	
	if ((s = semget(qkey, 2, 0666)) == -1)
		return -1;
	return semop(s, s_decrelease, 2);
}

int sem_wait4_msg_remove_it_then_block(key_t qkey)
{
	int s;
	
	if ((s = semget(qkey, 0, 0666)) == -1)
		return -1;
	return semop(s, s_wait4, 3);
}

int sem_release_queue(key_t qkey)
{
	int s;
	
	if ((s = semget(qkey, 2, 0666)) == -1)
		return -1;
	return semop(s, s_unlock, 1);
}

int sem_release_queue_rmsem(key_t qkey)
{
	int s;
	
	if ((s = semget(qkey, 2, 0666)) == -1)
		return -1;
	return semctl(s, 0, IPC_RMID, 0);
}

/*
 * informacje na temat kolejek otwartych w danym procesie
 */
static struct my_queue_info {
	int id;
	struct my_msqid_ds *seg;
} _my_qinfo[50];

int _my_add_queue(int id, key_t key, struct my_msqid_ds *seg)
{
	int i, ok = 0;

	for (i = 0; i < MY_QMAX; i++) {
		if (_my_qinfo[i].id == 0) {
			_my_qinfo[i].id = id;
			_my_qinfo[i].seg = seg;
			ok = 1;
			break;
		}
	}
	if (!ok) {
		return 0;
	}
	return 1;
}

int _my_del_queue(int id)
{
	int i, ok = 0;

	for (i = 0; i < MY_QMAX; i++) {
		if (_my_qinfo[i].id == id) {
			_my_qinfo[i].id = 0;
			_my_qinfo[i].seg = 0;
			ok = 1;
			break;
		}
	}
	if (!ok) {
		return 0;
	}
	return 1;
}

struct my_msqid_ds *_my_get_queue(int id)
{
	int i;

	for (i = 0; i < MY_QMAX; i++) {
		if (_my_qinfo[i].id == id)
			return _my_qinfo[i].seg;
	}
	return NULL;
}

struct my_msg *_my_shmalloc(size_t size, key_t startkey, key_t *endkey)
{
	struct my_msg *seg;
	int id, r = 100;
	key_t key = startkey + 1;
	while ((id = shmget(key, size, (IPC_CREAT | IPC_EXCL | 0666))) == -1) {
		if (errno == EEXIST) {
			if (key == startkey) {
				return NULL;
			} else {
				key++;
				if (key < 0) key = 100;
				if (key == startkey)
					key++;
			}
		} else {
			fprintf(stderr, "Failed to get a segment for the message\n");
			if (r) {
				fprintf(stderr, "Will try %d more times\n", r);
				r--;
			} else {
				return NULL;
			}
		}
	}
	if ((seg = shmat(id, 0, 0)) == NULL)
		return NULL;
	*endkey = key;
	return seg;
}

int _my_attach_msg(struct my_msqid_ds *where, key_t what, size_t size)
{
	int i = where->msg_begin;
	
	if (what == 0) {
		fprintf(stderr, "*** zero message key ***\n");
		exit(EXIT_FAILURE);
	}
	if (i == NO_MSGS) {
		i = 0;
		where->msg_begin = i;
		where->msgkeys[i] = what;
		where->msgsizes[i] = size;
	} else {
		for (i = ((i+1) % MY_MMAX); i != where->msg_begin; i = ((i+1) % MY_MMAX)) {
			if (where->msgkeys[i] == 0) {
				where->msgkeys[i] = what;
				where->msgsizes[i] = size;
				break;
			}
		}
	}
/*	fprintf(stderr, "Added msg %x at %d\n", what, i);*/
	fflush(stderr);
	where->msg_count++;
	return 1;
}

struct my_msg *_my_detach_msg(struct my_msqid_ds *where, int *sid)
{
	int i, id, ok =  0;
	key_t segkey;
	size_t size;
	if (where->msg_begin == NO_MSGS) {
		fprintf(stderr, "NO_MSGS\n");
		return NULL;
	}
	segkey = where->msgkeys[where->msg_begin];
	size   = where->msgsizes[where->msg_begin];
	where->msgkeys[where->msg_begin] = 0;
	where->msgsizes[where->msg_begin] = 0;
	for (i = ((where->msg_begin + 1) % MY_MMAX); i != where->msg_begin; i = (i+1) % MY_MMAX) {
		if (where->msgkeys[i] != 0) {
			where->msg_begin = i;
			ok = 1;
			break;
		}
	}
	if (!ok)
		where->msg_begin = NO_MSGS;
	if ((id = shmget(segkey, size, 0666)) == -1)
		return NULL;
	*sid = id;
/*	fprintf(stderr, "Removed msg %x from %d\n", segkey, i);*/
	fflush(stderr);
	where->msg_count--;
	return shmat(id, 0, 0);
}

int shm_msgget(key_t key, int msgflag)
{
	int id, sav_errno;
	struct my_msqid_ds *seg;

	if (sem_block_queue(key) == -1) {
		fprintf(stderr, "Could not block msgq for msgget\n");
		return -1;
	}
	if ((id = shmget(key, sizeof(struct my_msqid_ds), msgflag|IPC_CREAT|IPC_EXCL)) == -1) {
		/* TODO: nie wiadomo, czy flagi sa zgodne */
		sav_errno = errno;
		if (errno == EEXIST) {	/* juz jest */
			if ((msgflag & IPC_CREAT) && (msgflag & IPC_EXCL)) {
				/* ale chcielismy ja tylko dla siebie */
				sem_release_queue(key);
				errno = EEXIST;
				return -1;
			} else { /* i chcielismy ja tylko otworzyc */
				if ((id = shmget(key, sizeof(struct my_msqid_ds), msgflag)) == -1) {
					sem_release_queue(key);
					errno = ENOENT;
					return -1;
				}
				/* dodajemy ja sobie i zwracamy */
				seg = (struct my_msqid_ds *) shmat(id, 0, 0);
			}
		} else {		/* inny blad */
			sem_release_queue(key);
			errno = sav_errno;
			return id;
		}
	} else	{	/* nie bylo takiej, wlasnie ja utworzylismy */
		if (msgflag & IPC_CREAT) { /* chcielismy utworzyc */
			seg = (struct my_msqid_ds *) shmat(id, 0, 0);
			seg->msg_count = 0;
			seg->msg_begin = NO_MSGS;
			memset(seg->msgkeys, 0, MY_MMAX*sizeof(key_t));
			memset(seg->msgsizes, 0, MY_MMAX*sizeof(size_t));
			seg->key = key;
		} else {	/* ale niepowinnismy byli */
			sem_release_queue(key);
			errno = ENOENT;
			return -1;
		}
	}
	if (!_my_add_queue(id, key, seg)) {
		shmdt((void *)(seg));
		shmctl(id, IPC_RMID, NULL);
		sem_release_queue(key);
		errno = ENOSPC;
		return -1;
	}
	sem_release_queue(key);
	return id;
}

int shm_msgsnd(int msqid, struct my_msgbuf *ptr, int length, int flag)
{
	struct my_msqid_ds *mainseg;
	struct my_msg *msg;
	key_t msgkey;

	if ((mainseg = _my_get_queue(msqid)) == NULL)
		return -1;
	if ((msg = _my_shmalloc(length, mainseg->key, &msgkey)) == NULL)
		return -1;
	if (sem_block_queue(mainseg->key) == -1)
		return -1;
/* BLOCK */
	_my_attach_msg(mainseg, msgkey, length);
	memcpy(msg, ptr, length);
	shmdt((void *)(msg));
/* RELEASE */
	if (sem_inc_msg_release_queue(mainseg->key) == -1)
		return -1;
	return 0;
}

int shm_msgrcv(int msqid, struct my_msgbuf *ptr, int length, long msgtype, int flag)
{
	struct my_msqid_ds *mainseg;
	struct my_msg *msg;
	int msgid;

	if ((mainseg = _my_get_queue(msqid)) == NULL) {
		fprintf(stderr, "Could not get_queue\n");
		return -1;
	}
	if (sem_wait4_msg_remove_it_then_block(mainseg->key) == -1){
		fprintf(stderr, "Could not wait4\n");
		return -1;
	}
/* BLOCK */
	if ((msg = _my_detach_msg(mainseg, &msgid)) == NULL) {
		fprintf(stderr, "Could not detach\n");
		sem_release_queue(mainseg->key);
		return -1;
	}
	if (sem_release_queue(mainseg->key) == -1) {
		fprintf(stderr, "Could not release Q\n");
		return -1;
	}
/* RELEASE */
	memcpy(ptr, msg, length);
	if (shmdt((void *)(msg)) == -1) {
		fprintf(stderr, "Could not shmdt\n");
		return -1;
	}
	if (shmctl(msgid, IPC_RMID, NULL) == -1) {
		fprintf(stderr, "Could not IPC_RMID\n");
		return -1;
	}
	return length;
}

int shm_msgctl(int msqid, int cmd, struct my_msqid_ds *buff)
{
	struct my_msqid_ds *mainseg;
	key_t k;

	switch (cmd) {
	case IPC_RMID:
		if ((mainseg = _my_get_queue(msqid)) == NULL)
			return -1;
		if (sem_block_queue(k = mainseg->key) == -1)
			return -1;
		if (shmdt((void *)(mainseg)) == -1)
			return -1;
		if (shmctl(msqid, IPC_RMID, NULL) == -1)
			return -1;
		if (sem_release_queue_rmsem(k) == -1)
			return -1;
		break;
	default:
		return 0;
	}
	return 0;
}
